function [J , grad] = nnCostFunction(nn_params, input_layer_size, hidden_layer_size, num_labels, X, y, lambda)

%   NNCOSTFUNCTION Implements the neural network cost function for a TWO layer
%   neural network which performs classification
%   [J grad] = NNCOSTFUNCTON(nn_params, hidden_layer_size, num_labels, X, y, lambda) 
%   computes the cost and gradient of the neural network.
%   The parameters for the neural network are "unrolled" into the vector
%   nn_params and need to be converted back into the weight matrices. 
%   The returned parameter grad should be a "unrolled" vector of the
%   partial derivatives of the neural network.

% Initialisations

Theta1 = reshape(nn_params(1:hidden_layer_size * (input_layer_size + 1)), ...
                 hidden_layer_size, (input_layer_size + 1)); % (n1 , n+1)
Theta2 = reshape(nn_params((1 + (hidden_layer_size * (input_layer_size + 1))):end), ...
                 num_labels, (hidden_layer_size + 1)); % (n2 , n1+1) & n2 = num_labels
m = size(X,1);
X = [ones(m, 1) X]; % (m , n+1)
Y = zeros(num_labels,m);
for i = 1:m
    Y(y(i),i) = 1;
end

% Feedforward the neural network and return the cost in the variable J

Z_2 = Theta1*X';
A_2 = sigmoid(Z_2); % (n1 , m)
A_2 = [ones(size(A_2',1), 1) A_2']; % (m , n1+1)
A_3 = sigmoid(Theta2*A_2'); % (n2 , m)
J = -1/m*sum(sum(Y.*log(A_3) + (ones(num_labels,m)-Y).*log(ones(num_labels,m) - A_3)));
J = J + (lambda/2/m)*( sum(sum(Theta1(:,2:end).*Theta1(:,2:end))) + sum(sum(Theta2(:,2:end).*Theta2(:,2:end))) );

% Backpropagation algorithm to compute the gradients Theta1_grad and Theta2_grad
 
delta_3 = A_3 - Y; % (n2 , m)
Delta_2 = delta_3*A_2; % (n2 , n1+1)
delta_2 = (Theta2(:,2:end)'*delta_3).*sigmoidGradient(Z_2); % (n1 , m)
Delta_1 = delta_2*X; % (n1 , n+1)

Theta1_grad = Delta_1/m; % (n1 , n+1)
Theta1_grad(:,2:end) = Theta1_grad(:,2:end) + (lambda/m)*Theta1(:,2:end);
Theta2_grad = Delta_2/m; % (n2 , n+1)
Theta2_grad(:,2:end) = Theta2_grad(:,2:end) + (lambda/m)*Theta2(:,2:end);
grad = [Theta1_grad(:) ; Theta2_grad(:)];

end
